#!/usr/bin/env python

# Copyright 2017 Informatics Matters Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import multiprocessing
import subprocess
import tempfile
import threading
from multiprocessing.dummy import Pool as ThreadPool
import datetime

from rdkit import Chem

from pipelines_utils import parameter_utils, utils
from pipelines_utils_rdkit import rdkit_utils

lock = threading.Lock()
PDB_PATH = ""
WRITER = ""
COUNTER = 0
SUCCESS = 0
THRESHOLD = None


def run_and_get_ans(mol, pdb_path):
    global PDB_PATH
    smogmol = tempfile.NamedTemporaryFile("w",suffix=".sdf",delete=False).name
    utils.log("PDB: " + PDB_PATH + " ligand: " + smogmol)
    out_f = open(smogmol, "w")
    out_f.write(Chem.MolToMolBlock(mol))
    out_f.close()
    # Run command
    pli_path = "/usr/local/pli/bin/pli"
    cmd = [pli_path,"-protein",pdb_path,"-ligand",smogmol,"-mode","score","-output","system,scores","-exact_voronoi_areas","0","-selection","ligand","-oformat","json","-minimise","1","-warnings","0","-min_max_iter", "10"]
    utils.log("PLI CMD: " + " ".join(cmd))
    proc = subprocess.Popen(cmd,stdout=subprocess.PIPE)
    # Parse the output
    me = proc.stdout.read()
    if not me:
        return None
    return json.loads(me)

def run_dock(mol):
    global WRITER,COUNTER,SUCCESS,THRESHOLD
    answer_dict = run_and_get_ans(mol, PDB_PATH)
    COUNTER += 1
    if not answer_dict:
        utils.log("FAILED MOL", Chem.MolToSmiles(mol))
        return
    if THRESHOLD is not None:
        if answer_dict["system"]["pliff_score"] > THRESHOLD:
            return
    for ans in answer_dict["system"]:
        if ans.startswith(u"pliff"):
            mol.SetDoubleProp(str(ans), answer_dict["system"][ans])
    utils.log("SCORED MOL:", Chem.MolToSmiles(mol), answer_dict)
    lock.acquire()
    WRITER.write(mol)
    SUCCESS+=1
    WRITER.flush()
    lock.release()

def main():
    global PDB_PATH,WRITER,THRESHOLD
    parser = argparse.ArgumentParser(description='PLI scoring - Docking calculation.')
    parameter_utils.add_default_io_args(parser)
    parser.add_argument('--no-gzip', action='store_true', help='Do not compress the output (STDOUT is never compressed')
    parser.add_argument('-pdb', '--pdb_file', help="PDB file for scoring")
    parser.add_argument('-t', '--threshold', type=float, help="The maximum score to allow", default=None)
    parser.add_argument('--threads', type=int, help="Number of threads to used. Default is the number of cores", default=None)
    parser.add_argument('--thin', action='store_true', help='Thin output mode')

    args = parser.parse_args()
    utils.log("PLI Args: ", args)

    # Open up the input file
    input, suppl = rdkit_utils.default_open_input(args.input, args.informat)
    # Open the output file
    s_now = datetime.datetime.utcnow().strftime("%d-%b-%Y %H:%M:%S UTC")
    source = 'pipelines/docking/plip.py'
    output, WRITER, output_base = \
        rdkit_utils.default_open_output(args.output, "plip", args.outformat,
                                        compress=not args.no_gzip,
                                        thinOutput=args.thin,
                                        valueClassMappings={'pliff_iscore': 'java.lang.Float',
                                                            'pliff_cscore': 'java.lang.Float',
                                                            'pliff_nb_score': 'java.lang.Float',
                                                            'pliff_gscore': 'java.lang.Float',
                                                            'pliff_score': 'java.lang.Float',
                                                            'pliff_tscore': 'java.lang.Float'},
                                        datasetMetaProps={'created': s_now,
                                                          'source': source,
                                                          'description': 'PLI scoring of docked structures'}
                                        )

    PDB_PATH = args.pdb_file
    if args.threshold:
        THRESHOLD = args.threshold

    # Iterate over the molecules
    # WARNING - if using parallel processing the order of molecules is not preserved. Set args.threads to 1 to ensure this.
    pool = ThreadPool(args.threads if args.threads is not None else multiprocessing.cpu_count())
    pool.map(run_dock, suppl)
    pool.close()
    pool.join()
    # Close the file
    WRITER.close()

    if args.meta:
        utils.write_metrics(output_base, {'__InputCount__': COUNTER, '__OutputCount__': SUCCESS, 'PLI': COUNTER})

if __name__ == "__main__":
    main()
